package de.upb.infexba.ml.feature;

import java.io.File;
import java.io.FileOutputStream;
import java.io.OutputStreamWriter;
import java.io.Writer;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Date;
import java.util.List;
import java.util.TimeZone;

import weka.core.Instances;
import weka.core.SelectedTag;
import weka.core.converters.ConverterUtils.DataSource;
import weka.filters.unsupervised.attribute.Remove;
import weka.classifiers.Evaluation;
import weka.classifiers.meta.FilteredClassifier;
import weka.classifiers.functions.LibSVM;

/**
 * Helps to optimize the cost-parameter of the libSVM.
 * 
 * After choosing files for training, validation and testing, remover-options (to remove certain features from the files) and cost parameters that shall be
 * tested, it evaluates the parameters on the validation files, and finally tests the best parameter w.r.t the validation on the testing files and writes all
 * this results to a file.
 * 
 * @author kabu
 * 
 */
public class WekaCostOptimizer {

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// TODO: set folders and file names

	// NOTE: please insert testFiles in the following order: test-file, (transfer) validation-file, (transfer) test-file

	// music:

	 static String arffDir = "evaluation/Mittelwert Faktoren/";
	 static String dir = "music/";
	 static String trainingFile = "dataset_intention_05-25 11 40 59_music_split-training.arff";
	 static String validationFile = "dataset_intention_05-25 11 41 27_music_split-validation.arff";
	 static String[] testFiles = { "dataset_intention_05-25 11 41 54_music_split-test.arff",
	 "dataset_intention_05-25 11 43 44_smartphone_split-validation.arff", "dataset_intention_05-25 11 44 41_smartphone_split-test.arff" };

	// static String arffDir = "evaluation/mit selbstgewählten Faktoren/";
	// static String dir = "music/";
	// static String trainingFile = "dataset_intention_05-10 12 42 06_music-training.arff";
	// static String validationFile = "dataset_intention_05-10 12 42 32_music-validation.arff";
	// static String[] testFiles = { "dataset_intention_05-10 12 42 58_music-test.arff",
	// "dataset_intention_05-10 12 59 45_lfa-corpus-corpus-smartphone_split-validation.arff",
	// "dataset_intention_05-10 12 48 52_lfa-corpus-corpus-smartphone_split-test.arff" };

	// static String arffDir = "evaluation/berechnete Faktoren, mehr N-Gramme/";
	// static String dir = "music/";
	// static String trainingFile = "dataset_intention_05-24 10 38 08_music_split-training.arff";
	// static String validationFile = "dataset_intention_05-24 10 38 42_music_split-validation.arff";
	// static String[] testFiles = { "dataset_intention_05-24 10 39 11_music_split-test.arff",
	// "dataset_intention_05-24 10 54 46_smartphone_split-validation.arff",
	// "dataset_intention_05-24 10 55 47_smartphone_split-test.arff" };

	// smartphone:

//	static String arffDir = "evaluation/Mittelwert Faktoren/";
//	static String dir = "smartphone/";
//	static String trainingFile = "dataset_intention_05-25 11 49 00_smartphone_split-training.arff";
//	static String validationFile = "dataset_intention_05-25 11 49 22_smartphone_split-validation.arff";
//	static String[] testFiles = { "dataset_intention_05-25 11 49 44_smartphone_split-test.arff",
//			"dataset_intention_05-25 11 51 00_music_split-validation.arff", "dataset_intention_05-25 11 52 14_music_split-test.arff" };
//
//	 static String arffDir = "evaluation/berechnete Faktoren, mehr N-Gramme/";
//	 static String dir = "smartphone/";
//	 static String trainingFile = "dataset_intention_05-17 10 17 40_lfa-corpus-corpus-smartphone_split-training.arff";
//	 static String validationFile =
//	 "dataset_intention_05-17 10 18 02_lfa-corpus-corpus-smartphone_split-validation.arff";
//	 static String[] testFiles = { "dataset_intention_05-17 10 18 23_lfa-corpus-corpus-smartphone_split-test.arff",
//	 "dataset_intention_05-18 12 37 36_music-validation.arff", "dataset_intention_05-18 12 38 53_music-test.arff" };

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// TODO: set indices

	// NOTE: don't forget the "," at the end :P

	// music:
	static String index_pos = "23-67,1,";
	static String index_texttype = "21-22,3,2,";
	static String index_writingstyle = "335-382,90-148,193-304,20,4-6,7-15,";
	static String index_sentiment = "16-19,";
	static String index_trigramme = "68-89,149-192,";
	static String index_core = "305-334,";
	static String index_CLASS = "383,";

//	 // smartphone:
//	 static String index_pos = "23-67,1,";
//	 static String index_texttype = "21-22,3,2,";
//	 static String index_writingstyle = "130-192,259-373,411-462,20,4-6,7-15,";
//	 static String index_sentiment = "16,17-19,";
//	 static String index_trigramme = "68-129,193-258,";
//	 static String index_core = "374-410,";
//	 static String index_CLASS = "463,";

	static String[] indices = { index_pos, index_texttype, index_writingstyle, index_sentiment, index_trigramme, index_core };

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// TODO: set options

	// Do everything automatically (no further settings needed)
	static boolean automaticEvaluation = false;

	// Set the remover option
	static String remove = index_core;
	static boolean invertSelection = true; // index_CLASS will be added automatically

	// Do you want to use the getOptCostAndPctCorrect method?
	static boolean automaticSearch = true;

	// If you don't use the automaticSearch you can choose whether you want to use one of the specified
	// cost-arrays above or specify such an
	// cost-array by setting the from, to and stepwidth values.
	static boolean useArray = false;
	static double[] costs = WekaCostOptimizer.cost_large;
	static double from = 23;
	static double to = 25;
	static double stepwidth = 0.1;

	static double[] cost_onlyOne = { 1. };
	static double[] cost_standard = { 0.01, 0.05, 0.1, 0.5, 1., 2., 5., 10., 15., 20., 25. };
	static double[] cost_large = { 30., 35., 45., 50., 55., 60., 65., 70., 75., 80., 85. };
	static double[] cost_larger = { 90, 100, 110, 120, 130, 140, 150, 155, 160, 165, 170, 175, 180 };
	static double[] cost_largest = { 185, 190, 195, 200, 205, 210, 215, 220, 225, 230, 235, 240, 245, 250 };

	// ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	public static void main(String[] args) {
		try {

			// file that will contain an overview of all calculated optCosts
			SimpleDateFormat df = new SimpleDateFormat("MM-dd HH mm ss");
			df.setTimeZone(TimeZone.getDefault());
			File protocol = new File(arffDir + dir + df.format(new Date()) + "_costOpt.txt");
			Writer out = new OutputStreamWriter(new FileOutputStream(protocol));

			// if automaticEvaluation is selected, all remove options will be evaluated one after another
			int numberOfRuns = 1;
			if (automaticEvaluation) {
				remove = indices[0];
				invertSelection = true;
				automaticSearch = true;
				numberOfRuns = 11;
			}

			// evaluate the remove options and write results to file
			List<Double> results;
			for (int run = 0; run < numberOfRuns; run++) {

				// optimize cost parameter for current remove option
				results = evaluate();
				if (results.size() == 5)
					out.write(remove + "\t-> optC = " + round(results.get(0)) + " optPct = " + round(results.get(1)) + " test = " + round(results.get(2))
							+ " \ttransfer: valid = " + round(results.get(3)) + " test = " + round(results.get(4)) + "\n");
				else
					out.write("Error...");
				// next
				if (run <= 4)
					remove += indices[run + 1];
				else if( run < 10){
					remove = indices[run - 4];
					if(run == 5)
						out.write("\n");
				}
			}
			out.close();
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	private static List<Double> evaluate() {
		List<Double> results = new ArrayList<Double>();

		try {

			long time = System.currentTimeMillis();

			// set costs
			if (useArray)
				costs = getCosts(from, to, stepwidth);

			// removerOptions as String
			String v = invertSelection && !remove.isEmpty() ? "-V" : "";
			String c = invertSelection && !remove.isEmpty() ? index_CLASS : "";
			String removerOptions = remove + c;

			// result-file
			SimpleDateFormat df = new SimpleDateFormat("MM-dd HH mm ss");
			df.setTimeZone(TimeZone.getDefault());
			File protocol = new File(arffDir + dir + df.format(new Date()) + "_ remove = " + v + "-R" + removerOptions + "_costOpt.txt");
			Writer out = new OutputStreamWriter(new FileOutputStream(protocol));

			// print settings to console
			System.out.println("Remover-Options: " + v + "-R" + remove + c);
			System.out.print("Test-Mode: ");
			if (automaticSearch)
				System.out.println("automatic");
			else
				for (double i : costs)
					System.err.print(i + ", ");
			System.out.println("Write result to file: " + protocol.getAbsolutePath());

			// print settings to result-file
			out.write("Remover-Options: " + v + "-R" + removerOptions + "\n");
			out.write("Test: ");
			if (automaticSearch)
				out.write("automatisch :)\n\n");
			else
				for (double i : costs)
					out.write(i + ", ");
			out.write("\n");

			// get Instances for training
			DataSource trainSource = new DataSource(arffDir + dir + trainingFile);
			Instances training = trainSource.getDataSet();
			if (training.classIndex() == -1)
				training.setClassIndex(training.numAttributes() - 1);

			// get Instances for validation
			DataSource validationSource = new DataSource(arffDir + dir + validationFile);
			Instances validation = validationSource.getDataSet();
			if (validation.classIndex() == -1)
				validation.setClassIndex(validation.numAttributes() - 1);

			// search the optimal cost parameter
			double optCost = 0.;
			double optPctCorrect = 0.;
			if (automaticSearch) {
				double[] result = searchOptCostAndPctCorrect(training, validation, removerOptions, out);
				optCost = result[0];
				optPctCorrect = result[1];
			} else {
				double cost, tmpPctCorrect;
				for (int i = 0; i < costs.length; i++) {
					cost = costs[i];
					tmpPctCorrect = getPctCorrect(training, validation, removerOptions, cost, out);
					System.out.print("=> c = " + round(cost) + ":\t pctCorrect = " + round(tmpPctCorrect) + "\n");
					if (optPctCorrect < tmpPctCorrect) {
						optCost = cost;
						optPctCorrect = tmpPctCorrect;
					}
				}
			}

			// result
			results.add(optCost);
			results.add(optPctCorrect);
			out.write("\n=> OptC = " + round(optCost) + ":\t pctCorrect = " + round(optPctCorrect) + "\n\n");
			out.write("OptC evaluated on the testfiles:\n\n");

			// validate the found optimal cost parameter on all testFiles
			DataSource source;
			Instances instances;
			double pctCorrect = 0;
			for (String file : testFiles) {
				// get instances
				source = new DataSource(arffDir + dir + file);
				instances = source.getDataSet();
				if (instances.classIndex() == -1)
					instances.setClassIndex(instances.numAttributes() - 1);
				// evaluate the optCost parameter
				out.write(file + ":\n");
				// getPctCorrect(training, instances, removerOptions, 1., out);
				pctCorrect = getPctCorrect(training, instances, removerOptions, optCost, out);
				results.add(pctCorrect);
				out.write("\n");
			}

			// write some information to screen
			System.out.println("*********************************************************************************");
			System.out.println("\n\n=> OptC = " + round(optCost) + ":\t pctCorrect = " + round(optPctCorrect) + "\n");
			double elapsed = ((System.currentTimeMillis() - time) / 1000);
			System.out.println("\nTime elapsed: " + elapsed + " sec");

			// write down elapsed time
			out.write("\n\n(elapsed time = " + elapsed + ")");
			out.close();

		} catch (Exception e) {
			e.printStackTrace();
		}

		return results;
	}

	/**
	 * A simple search algorithm for the optimal cost with respect to the validation-sets.
	 * 
	 * @param removerOptions
	 * @param out
	 * @return
	 * @throws Exception
	 */
	private static double[] searchOptCostAndPctCorrect(Instances training, Instances validation, String removerOptions, Writer out) throws Exception {
		double[] costs;
		double optCost, optPctCorrect = 0;
		double tmpCost, tmpPctCorrect;
		double[] result;

		// start with standard costs
		costs = WekaCostOptimizer.cost_standard;
		result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
		optCost = result[0];
		optPctCorrect = result[1];
		// should we try larger values?
		if (optCost >= costs[costs.length - 2]) {
			costs = WekaCostOptimizer.cost_large;
			result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
			tmpCost = result[0];
			tmpPctCorrect = result[1];
			// found better pctCorrect value with a larger cost?
			if (tmpPctCorrect > optPctCorrect) {
				optCost = tmpCost;
				optPctCorrect = tmpPctCorrect;
				// should we try larger values?
				if (optCost >= costs[costs.length - 2]) {
					costs = WekaCostOptimizer.cost_larger;
					result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
					tmpCost = result[0];
					tmpPctCorrect = result[1];
					// found better pctCorrect value with a larger cost?
					if (tmpPctCorrect > optPctCorrect) {
						optCost = tmpCost;
						optPctCorrect = tmpPctCorrect;
						// should we try larger values?
						if (optCost >= costs[costs.length - 2]) {
							costs = WekaCostOptimizer.cost_largest;
							result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
							tmpCost = result[0];
							tmpPctCorrect = result[1];
							// found better pctCorrect value with a larger cost?
							if (tmpPctCorrect > optPctCorrect) {
								optCost = tmpCost;
								optPctCorrect = tmpPctCorrect;
							}
						}
					}
				}
			}
		}

		if (optCost > 1) {
			out.write("\n1-steps:\n");
			if (optCost >= 10) {
				costs = getCosts(optCost - 5, optCost + 5, 1);
			} else if (optCost == 5) {
				costs = getCosts(2, 9, 1);
			} else if (optCost == 2) {
				costs = getCosts(1, 5, 1);
			}
			result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
			optCost = result[0];
		}

		out.write("\n0.1-steps:\n");
		if (optCost > 1)
			costs = getCosts(optCost - 1, optCost + 1, 0.1);
		else
			costs = getCosts(0.1, 1.2, 0.1);
		result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
		optCost = result[0];

		out.write("\n0.01-steps:\n");
		if (optCost > 0.1)
			costs = getCosts(optCost - 0.1, optCost + 0.1, 0.01);
		else
			costs = getCosts(0.01, 0.12, 0.01);
		result = getOptCostAndOptPctCorrect(training, validation, removerOptions, costs, out);
		optCost = result[0];
		optPctCorrect = result[1];

		double[] output = { optCost, optPctCorrect };
		return output;
	}

	/**
	 * returns the optimal cost (out of a range of given costs) and the corresponding pctCorrect value with respect to a certain set of instances.
	 * 
	 * @param instances
	 * @param removerOptions
	 * @param costs
	 * @param out
	 * @return
	 * @throws Exception
	 */
	private static double[] getOptCostAndOptPctCorrect(Instances training, Instances validation, String removerOptions, double[] costs, Writer out)
			throws Exception {
		double cost, tmpPctCorrect;
		double optCost = 1.;
		double optPctCorrect = 0.;
		for (int i = 0; i < costs.length; i++) {
			cost = costs[i];
			tmpPctCorrect = getPctCorrect(training, validation, removerOptions, cost, out);
			if (optPctCorrect < tmpPctCorrect) {
				optPctCorrect = tmpPctCorrect;
				optCost = cost;
			}
		}
		double[] output = { optCost, optPctCorrect };
		return output;
	}

	/**
	 * @param instances
	 * @param removeAttributes
	 * @param cost
	 * @param out
	 * @return
	 * @throws Exception
	 */
	private static double getPctCorrect(Instances training, Instances validation, String removeAttributes, double cost, Writer out) throws Exception {
		double pctCorrect = 0;

		LibSVM libsvm;
		Remove remover;
		FilteredClassifier filteredClassifier;
		Evaluation eval;

		libsvm = new LibSVM();
		libsvm.setCost(cost);
		libsvm.setKernelType(new SelectedTag(LibSVM.KERNELTYPE_LINEAR, LibSVM.TAGS_KERNELTYPE));
		libsvm.setSVMType(new SelectedTag(LibSVM.SVMTYPE_C_SVC, LibSVM.TAGS_SVMTYPE));

		remover = new Remove();
		remover.setAttributeIndices(removeAttributes);
		remover.setInvertSelection(invertSelection);
		remover.setInputFormat(training); // inform filter about dataset **AFTER** setting options

		filteredClassifier = new FilteredClassifier();
		filteredClassifier.setFilter(remover);
		filteredClassifier.setClassifier(libsvm);

		filteredClassifier.buildClassifier(training);

		eval = new Evaluation(training);
		eval.evaluateModel(filteredClassifier, validation);
		pctCorrect = eval.pctCorrect();

		out.write("c = " + round(cost) + ":\t pctCorrect = " + round(pctCorrect) + "\n");
		System.out.print("c = " + round(cost) + ":\t pctCorrect = " + round(pctCorrect) + "\n");

		return pctCorrect;
	}

	private static double[] getCosts(double from, double to, double stepwidth) {
		double[] costs;
		ArrayList<Double> list = new ArrayList<Double>();
		for (double value = from; value <= to; value = value + stepwidth)
			list.add(round(value));
		costs = new double[list.size()];
		for (int i = 0; i < list.size(); i++)
			costs[i] = list.get(i);
		return costs;
	}

	private static double round(double number) {
		return Math.round(number * 10000) / 10000.0;
	}
}
